/* zocobj.c   ..  ZTF Operations (Object only option) -tab 27sep2017  04feb2022 */

/* version for mac os   -tab  05feb2022 */

#include "aux.h"
#include "mua.h"
#include "galactic.h"

/* Maximum diagonal radius. */
#define ZTFRADIUS 5.217989


/* ZFDB: ZTF Field Data (basic data: field ID,ra,dec) */
#define MAXZFDB 4000
struct ZFDBstructure {
  int id;
  double ra;
  double dec;
  int DS;
};
typedef struct ZFDBstructure ZFDBtype;


/* ZTF Field IDs. */
#define MAXZFI 4000
struct ZFIstructure {
  int num;          /* ZFI[0].num is number of entries. */
  int id;           /* ZTF Field ID number */
  int slice;        /* DEC slice entry number in ZFS[]. */
  double ra;        /* RA */
  double dec;       /* Declination */
};
typedef struct ZFIstructure ZFItype;


/* ZTF CCD Parameters. */
struct ZCPstructure {
  double xccd[16][4];
  double yccd[16][4];
  double xave[16][4];    /* only [][0] and [][2] used */
  double yave[16][4];    /* only [][0] and [][2] used */
};
typedef struct ZCPstructure ZCPtype;


/* ZTF ccd POsition results for a given ra,dec */
struct ZPOstructure {
  int num;
  int index;
  int field;
  int ccd;
  double xfrac;   /* fractional position within given CCD */
  double yfrac;
  double xpos;    /* position relative to field center in degrees */
  double ypos;
  double sep;     /* distance between field center and object in degrees */
};
typedef struct ZPOstructure ZPOtype;




/* ----------------------------------------------------------------------
 * 333
 Grab a parameter.  Return a double precision value.
*/
double zoc_grab_parameter8( char line[], char parm[], double def )
{
/**/
int ii,jj,kk;
char wrd[22000];
double val,rr;
/**/
val = def;
ii = cindex(line,parm);
kk = clc(parm);
if (ii > 0) {
  substrcpy_terminate(line,ii+kk+1,clc(line),wrd,0); 
  jj = cindex(wrd,"&"); 
  if (jj <= 0) { rr=def; } else { wrd[jj]='\0'; rr=GLVQ(wrd,1); }
  if (rr > -1.e+39) { val = rr; }
}
return(val);
}


/* ----------------------------------------------------------------------
 * 333
 Grab a parameter.  Returns a string in 'val'.
*/
void zoc_grab_parameter( char line[], char parm[], char val[], char def[] )
{
/**/
int ii,jj,kk;
char wrd[22000];
/**/
strcpy(val,def);
ii = cindex(line,parm);
kk = clc(parm);
if (ii > 0) {
  substrcpy_terminate(line,ii+kk+1,clc(line),wrd,0); 
  jj = cindex(wrd,"&"); 
  if (jj <= 0) { strcpy(wrd,def); } else { wrd[jj]='\0'; }
  strcpy(val,wrd);
}
return;
}



/* ----------------------------------------------------------------------
 * 333
 Load ZCP.
*/
void zoc_load_ZCP( ZCPtype ZCP[] )
{
/**/
int ii,jj;
FILE *infu;
char line[100];
/**/

/* Center between 6 and 9 is at 0.,0. arcseconds.  */
/* CCD layout? :
   0  1  2  3
   4  5  6  7
   8  9 10 11
  12 13 14 15   */
/* CCD corners: 1 2
                0 3  */

/* Read numbers. */
if (FileExist("ztf_corners.tbl")==0) {
  printf("***error: could not find ztf_corners.tbl.\n");
  return;
}
infu = fopen_read("ztf_corners.tbl");
fgetline(line,infu);
for (ii=0; ii<16; ++ii) {
  for (jj=0; jj<4; ++jj) {
    fgetline(line,infu);
    ZCP[0].xccd[ii][jj] = GLV(line,1);
    ZCP[0].yccd[ii][jj] = GLV(line,2);
    ZCP[0].xave[ii][jj] = 0.;
    ZCP[0].yave[ii][jj] = 0.;
  }
  ZCP[0].xave[ii][0] = (ZCP[0].xccd[ii][0] + ZCP[0].xccd[ii][1]) / 2.;
  ZCP[0].xave[ii][2] = (ZCP[0].xccd[ii][2] + ZCP[0].xccd[ii][3]) / 2.;
  ZCP[0].yave[ii][0] = (ZCP[0].yccd[ii][0] + ZCP[0].yccd[ii][3]) / 2.;
  ZCP[0].yave[ii][2] = (ZCP[0].yccd[ii][1] + ZCP[0].yccd[ii][2]) / 2.;
}
fclose(infu);

return;
}


/* ----------------------------------------------------------------------
 * 333
  Find all CCDs that a given ra,dec lands on and return field and CCD position list.
  Also includes ZPO[] entry for closest field (may be in a gap).
*/
void zoc_ztf_ccdpos( double ra, double dec, int grid, ZFItype ZFI[], ZCPtype ZCP, ZPOtype ZPO[] )
{
/**/
int ok,ii,loii1,loii2,kk,nn;
double sep,x,y,z,xp,yp,zp,xpos,ypos;
double loxpos1,loypos1,losep1;
double loxpos2,loypos2,losep2;
RMtype RM;
/**/
/* Search all fields (with ID < 2000). */
nn=0; 
loii1=-1; loxpos1=0.; loypos1=0.; losep1=9999.;
loii2=-1; loxpos2=0.; loypos2=0.; losep2=9999.;
for (ii=0; ii<ZFI[0].num; ++ii) {
if (ZFI[ii].id < 2000) {
  ok=0;
  if (grid == 1) { if (ZFI[ii].id < 1000) ok=1; } else {
  if (grid == 2) { if (ZFI[ii].id > 1000) ok=1; } else {
                 {                        ok=1; }      }}
  if (ok) {
    sep = cangsep(ra,dec,ZFI[ii].ra,ZFI[ii].dec);
    if (sep < ZTFRADIUS) {
      SetRotationMatrix( degtorad(ZFI[ii].ra), degtorad(ZFI[ii].dec), &RM );
      radec2xyz(ra,dec,&x,&y,&z);
      xyz2prime(x,y,z,RM.mat,&xp,&yp,&zp);
      xpos = radtodeg(yp);
      ypos = radtodeg(zp);
/* Closest field each grid. */
      if (ZFI[ii].id < 1000) {
        if (sep < losep1) { losep1=sep; loii1=ii; loxpos1=xpos; loypos1=ypos; }
      } else {
        if (sep < losep2) { losep2=sep; loii2=ii; loxpos2=xpos; loypos2=ypos; }
      }
/* Fields for which position is on a CCD. */
      for (kk=0; kk<16; ++kk) {
        if ((xpos > ZCP.xave[kk][0])&&(xpos < ZCP.xave[kk][2])&&
            (ypos > ZCP.yave[kk][0])&&(ypos < ZCP.yave[kk][2])) {
          ZPO[nn].index = ii;
          ZPO[nn].field = ZFI[ii].id;
          ZPO[nn].ccd   = kk;
          ZPO[nn].xfrac = (xpos - ZCP.xave[kk][0]) / (ZCP.xave[kk][2] - ZCP.xave[kk][0]);
          ZPO[nn].yfrac = (ypos - ZCP.yave[kk][0]) / (ZCP.yave[kk][2] - ZCP.yave[kk][0]);
          ZPO[nn].xpos  = xpos;
          ZPO[nn].ypos  = ypos;
          ZPO[nn].sep   = sep;
          ++nn;
        }
      }
    }
  }
}
}
if (loii1 > -1) {
  ZPO[nn].index = loii1;
  ZPO[nn].field = ZFI[loii1].id;
  ZPO[nn].ccd   = -1;
  ZPO[nn].xfrac = -9.;
  ZPO[nn].yfrac = -9.;
  ZPO[nn].xpos  = loxpos1;
  ZPO[nn].ypos  = loypos1;
  ZPO[nn].sep   = losep1;
  ++nn;
}
if (loii2 > -1) {
  ZPO[nn].index = loii2;
  ZPO[nn].field = ZFI[loii2].id;
  ZPO[nn].ccd   = -1;
  ZPO[nn].xfrac = -9.;
  ZPO[nn].yfrac = -9.;
  ZPO[nn].xpos  = loxpos2;
  ZPO[nn].ypos  = loypos2;
  ZPO[nn].sep   = losep2;
  ++nn;
}
ZPO[0].num = nn;
return;
}

/*
      printf("xave0=%8.4f  xave2=%8.4f  yave0=%8.4f  yave2=%8.4f\n", ZCP.xave[kk][0], ZCP.xave[kk][2], ZCP.yave[kk][0], ZCP.yave[kk][2]);
*/




/* ----------------------------------------------------------------------
 * 333
 Convert Tom's CCD# to IPAC ZTF CCD number.
*/
int zoc_IPAC_ccdno( int ccdno )
{
/**/
int ii;
/**/

if (ccdno == 0) ii=16;
if (ccdno == 1) ii=15;
if (ccdno == 2) ii=14;
if (ccdno == 3) ii=13;

if (ccdno == 4) ii=12;
if (ccdno == 5) ii=11;
if (ccdno == 6) ii=10;
if (ccdno == 7) ii= 9;

if (ccdno == 8) ii= 8;
if (ccdno == 9) ii= 7;
if (ccdno ==10) ii= 6;
if (ccdno ==11) ii= 5;

if (ccdno ==12) ii= 4;
if (ccdno ==13) ii= 3;
if (ccdno ==14) ii= 2;
if (ccdno ==15) ii= 1;

return(ii);
}



/* ----------------------------------------------------------------------
 * 333
  Replace a code, return '1' is a replacement was made.
*/
int zoc_replace( char wrd[], char code[], char rep[] )
{
/**/
int ok,ii;
char wrda[200],wrdb[200];
/**/
ok=0;
ii=cindex(wrd,code);
if (ii > -1) {
  if (ii == 0        ) { strcpy(wrda,""); } else { substrcpy_terminate(wrd,0,ii-1,wrda,0);        }
  if (ii+3 > clc(wrd)) { strcpy(wrdb,""); } else { substrcpy_terminate(wrd,ii+3,clc(wrd),wrdb,0); }
  strcpy(wrd,wrda); strcat(wrd,rep); strcat(wrd,wrdb);
  ok=1;
}
return(ok);
}


/* ----------------------------------------------------------------------
 * 333
  Replace some select control codes in a string.
  Fix or clean up a string by replacing special characters.
*/
void zoc_replace_codes( char wrd[] )
{
/**/
int ii,again,count;
char dq[2];
/**/
dq[0]='"'; dq[1]='\0';
/* A plus means a space. */
for (ii=0; ii<clc(wrd); ++ii) { if (wrd[ii] == '+') wrd[ii]=' '; }
/* Replace. */
again=1; count=0;
while (again) {
  again=0;
  if (zoc_replace( wrd, "%2B", "+" )) again = 1;
  if (zoc_replace( wrd, "%3A", ":" )) again = 1;
  if (zoc_replace( wrd, "%21", "!" )) again = 1;
  if (zoc_replace( wrd, "%24", "$" )) again = 1;
  if (zoc_replace( wrd, "%5E", "^" )) again = 1;
  if (zoc_replace( wrd, "%26", "&" )) again = 1;
  if (zoc_replace( wrd, "%40", "@" )) again = 1;
  if (zoc_replace( wrd, "%23", "#" )) again = 1;
  if (zoc_replace( wrd, "%25", "%" )) again = 1;
  if (zoc_replace( wrd, "%28", "(" )) again = 1;
  if (zoc_replace( wrd, "%29", ")" )) again = 1;
  if (zoc_replace( wrd, "%3D", "=" )) again = 1;
  if (zoc_replace( wrd, "%5B", "[" )) again = 1;
  if (zoc_replace( wrd, "%5D", "]" )) again = 1;
  if (zoc_replace( wrd, "%7B", "{" )) again = 1;
  if (zoc_replace( wrd, "%7D", "}" )) again = 1;
  if (zoc_replace( wrd, "%7C", "|" )) again = 1;
  if (zoc_replace( wrd, "%3B", ";" )) again = 1;
  if (zoc_replace( wrd, "%22", dq  )) again = 1;
  if (zoc_replace( wrd, "%27", "'" )) again = 1;
  if (zoc_replace( wrd, "%3C", "<" )) again = 1;
  if (zoc_replace( wrd, "%3E", ">" )) again = 1;
  if (zoc_replace( wrd, "%2C", "," )) again = 1;
  if (zoc_replace( wrd, "%3F", "?" )) again = 1;
  if (zoc_replace( wrd, "%2F", "/" )) again = 1;
  ++count;
  if (count > 100) again=0;
}
return;
}






/* ----------------------------------------------------------------------
 * 333
  Version: manual (from command line)
*/
void zoc_show_object_manual( char QS[] )
{
/**/
int count,countgap,kk,ccdno,ee,ii,num,grid;
/**/
char wrd[100];
char rawrd[100];
char decwrd[100];
char object[100];
char gridwrd[10];
/**/
double objra,objdec;
/**/
ZPOtype ZPO[100];
ZFDBtype *ZFDB;
ZFItype *ZFI;
ZCPtype ZCP;
/**/
FILE *binfu;
/**/

/* Check and load ZFDB file. */
if (FileExist("ZFDB.bin") == 0) {
  printf("Content-type: text/plain\n\n");
  printf("***error: could not find ZFDB.bin file.\n");
  return;
}
ZFDB = (ZFDBtype *)calloc(MAXZFDB,sizeof(ZFDBtype));
binfu = fopen_read("ZFDB.bin");
fread((char *)&num, 1, sizeof(int), binfu);
for (ii=0; ii<num; ++ii) {
  fread((char *)&ZFDB[ii], 1, sizeof(ZFDBtype), binfu);
}
fclose(binfu);

/* Allocate and copy. */
ZFI = (ZFItype *)calloc(MAXZFI,sizeof(ZFItype));
for (ii=0; ii<num; ++ii) {
  ZFI[ii].id = ZFDB[ii].id;
  ZFI[ii].ra = ZFDB[ii].ra;
  ZFI[ii].dec= ZFDB[ii].dec;
  ZFI[ii].slice = ZFDB[ii].DS;
}
ZFI[0].num = num;


/* Load corners data from 'ztf_corners.tbl'. */
zoc_load_ZCP( &ZCP );

zoc_grab_parameter(QS,"objra=",rawrd,"");
zoc_replace_codes( rawrd );
if (cindex(rawrd,":") > 0) { objra = misc_hmss2deci( rawrd, 1 ); } 
                      else { objra = GLV(rawrd,1); sprintf(rawrd,"%9.5f",objra); }

zoc_grab_parameter(QS,"objdec=",decwrd,"");
zoc_replace_codes( decwrd );
if (cindex(decwrd,":") > 0) { objdec = misc_hmss2deci( decwrd, 0 ); } 
                       else { objdec = GLV(decwrd,1); sprintf(decwrd,"%9.5f",objdec); }

grid = zoc_grab_parameter8(QS,"grid=",1.);

zoc_grab_parameter(QS,"objname=",object,"unknown");
zoc_replace_codes( object );

ZPO[0].num=-1;

if ((grid == 1)||(grid == 2)) { sprintf(gridwrd,"%d",grid); }
                         else { strcpy(gridwrd,"1 or 2");   }

zoc_ztf_ccdpos( objra, objdec, grid, ZFI, ZCP, ZPO );

misc_pad_blanks( object, 30, wrd );
count=0; countgap=0;
for (ii=0; ii<ZPO[0].num; ++ii) {  
  if (ZPO[ii].ccd > -1) {
    ee=ZPO[ii].index;
    ccdno = zoc_IPAC_ccdno( ZPO[ii].ccd );
    kk=1;
    if (ZPO[ii].field > 1000) kk=2;
    printf(" %6d %4d %3d %6.3f %6.3f %8.4f %8.4f %6.3f %s %8.4f %8.4f \n",
      ZPO[ii].field,kk,ccdno,ZPO[ii].xfrac,ZPO[ii].yfrac,ZFI[ee].ra,ZFI[ee].dec,ZPO[ii].sep,wrd,objra,objdec);
    ++count;
  } else {
    ++countgap;
  }
}
if (count == 0) {
  for (ii=0; ii<ZPO[0].num; ++ii) {  if (ZPO[ii].ccd == -1) {
    ee=ZPO[ii].index;
    kk=1;
    if (ZPO[ii].field > 1000) kk=2;
    printf(" %6d %4d %3d %6.3f %6.3f %8.4f %8.4f %6.3f %s %8.4f %8.4f \n",
      ZPO[ii].field,kk,-1,ZPO[ii].xfrac,ZPO[ii].yfrac,ZFI[ee].ra,ZFI[ee].dec,ZPO[ii].sep,wrd,objra,objdec);
  }}
}

/* Exit in manual mode. */
free(ZFI);
free(ZFDB);
return;
}



/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* 333 */
int main(int argc, char *argv[])
{
/**/
char QS[22000];
/**/
char arg[202][200];
int narg;
/**/
int grid,ii1,ii2,ii3;
char wrd[200];
char objname[200];
char listfile[200];
char line[1000];
double ra,dec,rad;
FILE *infu;
/**/

/* Copy arguments. */
cargcopy(argv, argc, &narg, arg);
if (narg == 0) {
  printf("Syntax: zocobj  ( [ ra= dec= name= ]  or  [ list= ] )  [ rad= ] [ grid= ]\n");
  return(0);
}

/* Show object position on ZTF field grid. */
  ra=0.;  if (cfindarg(arg,&narg,"ra="  ,':',wrd)) { ra  =GLV(wrd,1); }
  dec=0.; if (cfindarg(arg,&narg,"dec=" ,':',wrd)) { dec =GLV(wrd,1); }
  rad=-1; if (cfindarg(arg,&narg,"rad=" ,':',wrd)) { rad =GLV(wrd,1); }
  grid=1; if (cfindarg(arg,&narg,"grid=",':',wrd)) { grid=GLV(wrd,1); }
  strcpy(objname,"unknown"); if (cfindarg(arg,&narg,"name=",':',wrd)) { strcpy(objname,wrd); }
  strcpy(listfile,""); if (cfindarg(arg,&narg,"list=",':',wrd)) { strcpy(listfile,wrd); }


/* Print header. */
    printf("|field |grid|ccd|xfrac |yfrac | ra     | dec    | sep  | object                       | objra  | objdec |\n");
/*           123456 1234 123 123456 123456 12345678 12345678 123456 123456789012345678901234567890 12345678 12345678 */

if ((ABS((ra)) > 0.0001)||(ABS((dec)) > 0.0001)) {
  sprintf(QS,"showobject=1&objra=%f&objdec=%f&grid=%d&objname=%s&radam=%f",ra,dec,grid,objname,rad);
  zoc_show_object_manual(QS);
  return(0);
} else {
  if (strcmp(listfile,"") == 0) {
    printf("Syntax:  zoc  list=(filename)  grid=(1,2,or 3(both))\n");
    printf(" ..or..\n");
    printf("Syntax:  zoc  ra=(ra in degrees)  dec=(dec in degrees)  grid=(1,2,or 3(both))\n");
    return(0);
  }
}
infu = fopen_read(listfile);
while (fgetline(line,infu)) {
  ii1 = cindex(line,",");
  substrcpy_terminate(line,0,ii1-1,objname,0);
  line[ii1]=' ';
  ii2 = cindex(line,",");
  substrcpy_terminate(line,ii1+1,ii2-1,wrd,0);
  ra = GLVQ(wrd,1);
  if ((ra > -1.)&&(ra < 361.)) {
    line[ii2]=' ';
    ii3 = cindex(line,",");
    if (ii3 < ii2) ii3=clc(line)+1;
    substrcpy_terminate(line,ii2+1,ii3-1,wrd,0);
    dec = GLVQ(wrd,1);
    if ((dec > -91.)&&(dec < 91.)) {
      sprintf(QS,"showobject=1&objra=%f&objdec=%f&grid=%d&objname=%s&radam=%f",ra,dec,grid,objname,rad);
      zoc_show_object_manual(QS);
    } else {
      printf("...Bad DEC on line: %s ...\n",line);
    }
/*
  } else {
    printf("...Bad RA on line %s ...\n",line);
*/
  }
}
fclose(infu);

return(0);
}

